@@ -1,8 +1,9 @@
 # Data::Hopen::G::Link - base class for hopen edges
 package Data::Hopen::G::Link;
+use strict;
 use Data::Hopen::Base;
 
-our $VERSION = '0.000014'; # TRIAL
+our $VERSION = '0.000016'; # TRIAL
 
 use parent 'Data::Hopen::G::Runnable';
 use Class::Tiny {
@@ -32,6 +33,9 @@ Copy the inputs to the outputs.
 
 The output is C<{}> if no inputs are provided.
 
+By default, the output is a copy of the inputs
+(see L<Data::Hopen::G::Runnable/passthrough>).
+
 =cut
 
 sub _run {
